@@ -1,6 +1,6 @@
 # High Availability package for DBI
 #
-# Copyright (c) 2003-2004 Henri Asseily <henri@asseily.com>. All rights reserved.
+# Copyright (c) 2003-2006 Henri Asseily <henri@asseily.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
@@ -9,9 +9,9 @@ package DBIx::HA;
 
 use 5.006000;
 
-use constant DBIx_HA_DEBUG => 0;
+use constant DBIx_HA_DEBUG => $ENV{DBIX_HA_DEBUG} || 0;
 use Data::Dumper;
-use DBI 1.44 ();
+use DBI 1.49 ();
 use Sys::SigAction qw( set_sig_handler );
 use Exporter ();
 use strict;
@@ -20,13 +20,20 @@ use vars qw ( @ISA $prefix );
 
 our $loaded_Apache = 0;
 our $loaded_Apache_DBI = 0;
+our $logdir;
 
 BEGIN {
-	$DBIx::HA::VERSION = 0.95;
+	$DBIx::HA::VERSION = 1.10;
+
+	# Sample Fail Test Functions for different Database Servers
+	# They are used in the configuration 
+	# Input is ($ErrorID, $ErrorString)
+	# Output is boolean: If true, then we'll consider the error a critical condition, ok to failover
+	# If false, then DBIx::HA will not act on it and pass it straight through to the client
+	sub FTF_SybaseASE { $_[0] > 10 ? 0 : 1 }; # Fail Test Function for Sybase ASE
 }
 
 our $prefix = "[$$] DBIx::HA:           "; 
-my $logdir;
 
 sub initialize {
 	if ($Apache::VERSION) {
@@ -51,15 +58,22 @@ sub initialize {
 		}
 		# add default timeouts for connection and execution
 		if (! $DATABASE::conf{$dbname}->{'connecttimeout'}) {
-			$DATABASE::conf{$dbname}->{'connecttimeout'} = 5;
+			$DATABASE::conf{$dbname}->{'connecttimeout'} = 30;
 		}
 		if (! $DATABASE::conf{$dbname}->{'executetimeout'}) {
-			$DATABASE::conf{$dbname}->{'executetimeout'} = 10;
+			$DATABASE::conf{$dbname}->{'executetimeout'} = 30;
+		}
+		# add default failover rule based on DBI error
+		# default is to not request failover, whatever the error
+		# do not ever set the default to always request failover, as any SQL error would trigger failover!
+		if (! $DATABASE::conf{$dbname}->{'failtest_function'}) {
+			$DATABASE::conf{$dbname}->{'failtest_function'} = sub { 0 };
 		}
 		$DATABASE::conf{$dbname}->{'end_of_stack'} = 0;	# error condition reaching end of stack
-		foreach (@{$DATABASE::conf{$dbname}->{'db_stack'}}) {
+		for my $conn_aref (@{$DATABASE::conf{$dbname}->{'db_stack'}}) {
+            my $dsn = $conn_aref->[0];
 			# create an easy reverse-lookup table for finding the db server name from the dsn
-			$DBIx::HA::finddbserver{$_->[0]}  = $dbname;
+			$DBIx::HA::finddbserver{$dsn}  = $dbname;
 			# add timeout when within Apache::DBI
 			# default to no ping (-1)
 			if ($loaded_Apache_DBI) {
@@ -67,8 +81,8 @@ sub initialize {
 					die "$prefix Requirement unmet. Apache::DBI must be at version 0.89 or above";
 				}
 				# create a cached lookup table for finding the Apache::DBI cache key index from the dsn
-				$DBIx::HA::ApacheDBIidx{$_->[0]}  = _getApacheDBIidx(@$_);
-				Apache::DBI->setPingTimeOut($_->[0], $DATABASE::conf{$dbname}->{'pingtimeout'} || -1);
+				$DBIx::HA::ApacheDBIidx{$dsn}  = _getApacheDBIidx(@$conn_aref);
+				Apache::DBI->setPingTimeOut($dsn, $DATABASE::conf{$dbname}->{'pingtimeout'} || -1);
 			}
 		};
 		# set the active database to be the first in the stack
@@ -95,7 +109,7 @@ sub _init_child {
 	foreach $dbname (keys %DATABASE::conf) {
 		# under application failover, maybe we already have an active db.
 		# set the active database to be the first in the stack unless we got it earlier.
-		_readsharedfile($dbname) unless ($loaded_Apache);
+		_readsharedfile($dbname);
 		_writesharedfile($dbname, 0) unless ($DATABASE::conf{$dbname}->{'active_db'});
 
 		# allow for connect on initialization
@@ -112,28 +126,27 @@ sub _readsharedfile {
 	my $dbname = shift;
 	if ($DATABASE::conf{$dbname}->{'failoverlevel'} eq 'application') {
 		# do this only if we're doing application failover and not process failover
-		if ($loaded_Apache) {
-			# $logdir = Apache::server_root_relative(undef,'logs'); # unnecessary since set during _writesharedfile on init
-			my $r = Apache->request;
-			if ($r && (! defined($r->notes("activedb_$dbname"))) && (-f "$logdir/DBIxHA_activedb_$dbname")) {
-				open IN, "$logdir/DBIxHA_activedb_$dbname";
-				my $dbidx = <IN>;
-				chomp $dbidx;
-				close IN;
-				if ($dbidx == -1) {	# we're told that we're at the end of the stack. No db server is available.
-					$DATABASE::conf{$dbname}->{'end_of_stack'} = 1;
-					return 0;
-				}
-				$DATABASE::conf{$dbname}->{'end_of_stack'} = 0;
-				if (($dbidx =~ /^\d+$/o) && $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx]) {
-					$DATABASE::conf{$dbname}->{'active_db'} = $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx];
-					$DBIx::HA::activeserver{$dbname}  = $dbidx;
-					$r->notes("activedb_$dbname", $dbidx);
-				} else {
-					warn "$prefix in _isactivedb: $dbname shared file has erroneous content, overwriting.\n";
-					_writesharedfile($dbname, $DBIx::HA::activeserver{$dbname});
-					return 0;
+		if ($loaded_Apache && (-f "$logdir/DBIxHA_activedb_$dbname")) {
+			open IN, "$logdir/DBIxHA_activedb_$dbname";
+			my $dbidx = <IN>;
+			chomp $dbidx;
+			close IN;
+			if ($dbidx == -1) {	# we're told that we're at the end of the stack. No db server is available.
+				$DATABASE::conf{$dbname}->{'end_of_stack'} = 1;
+				return 0;
+			}
+			$DATABASE::conf{$dbname}->{'end_of_stack'} = 0;
+			if (($dbidx =~ /^\d+$/o) && $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx]) {
+				$DATABASE::conf{$dbname}->{'active_db'} = $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx];
+				$DBIx::HA::activeserver{$dbname}  = $dbidx;
+				unless ($Apache::ServerStarting == 1) {
+					my $r = Apache->request;
+					$r->notes("activedb_$dbname", $dbidx) if (ref $r);
 				}
+			} else {
+				warn "$prefix in _readsharedfile: $dbname shared file has erroneous content, overwriting.\n";
+				_writesharedfile($dbname, $DBIx::HA::activeserver{$dbname});
+				return 0;
 			}
 		}
 	}
@@ -282,19 +295,24 @@ sub _reconnect {
 			$DATABASE::retries{$currdsn} = 0 if (! $DATABASE::retries{$currdsn});
 			for ($i=$DATABASE::retries{$currdsn}; $i < $DATABASE::conf{$dbname}->{'max_retries'}; $i++) {	# retry max_retries
 				$DATABASE::retries{$currdsn}++;
-				# now try to destroy, clear, undefine every instance and pointer of and to the $dbh
+				# now we're going to create a new good database handle and swap it with the old bad one
 				$newdbh = _connect_with_timeout (@$selrow);
 				if (defined $newdbh) {
-					# all is good
+					# we managed to create a new database handle
 					if ((defined $dbh) && (ref $dbh)) {
+						# the old one still exists, so we're going to swap it and then destroy it
 						warn "$prefix in _reconnect: Pointing dbh to newdbh\n" if (DBIx_HA_DEBUG);
 						$dbh->swap_inner_handle($newdbh);
-						eval { $newdbh->disconnect; };
-						undef $newdbh;
+						# wipe the old database handle (which in turn finishes all its children sth's)
+						# If we're using DBD::Sybase, make sure syb_flush_finish is off so we don't get remaining results
+                                                $newdbh->{syb_flush_finish} = 0 if $newdbh->{syb_flush_finish}
+                                                    and $newdbh->{Driver}{Name} ne 'Gofer';
+						eval { undef $newdbh; };
 					} else {
+						# there was no previous active database handle, so that's easy
 						$dbh = $newdbh;
 					}
-					warn "$prefix in _reconnect: connected to $currdsn\n" if (DBIx_HA_DEBUG);
+					warn "$prefix Successfully reconnected to $currdsn\n";
 					$DATABASE::retries{$currdsn} = 0; # reset the retries counter
 					_writesharedfile($dbname, $dbstackindex);
 					# Do callback if it exists
@@ -314,7 +332,7 @@ sub _reconnect {
 				warn "$prefix *** ERROR: Exhausted DBI failover stack. Last DSN is: $olddsn \n";
 				return ($olddsn, undef);
 			}
-			warn "$prefix in _reconnect: dbstackindex: $dbstackindex; Trying another db server: $currdsn \n" if (DBIx_HA_DEBUG);
+			warn "$prefix in _reconnect: dbstackindex: $dbstackindex; Trying another db server: $currdsn \n";
 			goto FINDDB;
 		} #if
 		$dbstackindex++;
@@ -329,7 +347,11 @@ sub connect {
 	warn Dumper Apache::DBI::all_handlers() if (DBIx_HA_DEBUG > 1 && $loaded_Apache_DBI);
 	my $class = shift;
 	my $dbname = shift;
-	my ($dsn, $username, $auth, $attrs) = @{$DATABASE::conf{$dbname}->{'active_db'}};
+	my $conf = $DATABASE::conf{$dbname}
+            or Carp::croak("No entry for '$dbname' in %DATABASE::conf");
+	my $active_db = $conf->{'active_db'}
+            or Carp::croak("No active_db for '$dbname' (did you call initialize?)");
+	my ($dsn, $username, $auth, $attrs) = @$active_db;
 
 	# Update the active db. If it's been updated, switch to it
 	if (! _isactivedb($dsn)) {
@@ -369,19 +391,21 @@ sub _connect_with_timeout {
 	alarm(0);
 	if ($@ or $timeout) {	# there's a problem above
 		if ($timeout) {	# it's a timeout
-			warn "$prefix CONNECT TIMED OUT in $dsn";
+			warn "$prefix *** CONNECT TIMED OUT in $dsn";
 			eval { $dbh->disconnect };
 			$dbh = undef;
 		} else {	# problem in the connection
 			warn "$prefix Error in DBI::connect: $@\n" if $@;
 		}
 	}
+        $dbh->{private_dbixha_dsn} = $dsn if $dbh;
 	return $dbh;
 }
 } # end package DBIx::HA
 
 {
 package DBIx::HA::db;
+use strict;
 use constant DBIx_HA_DEBUG => DBIx::HA::DBIx_HA_DEBUG;
 use vars qw ( @ISA );
 @ISA = qw(DBI::db DBIx::HA);
@@ -392,10 +416,15 @@ sub prepare {
 	my $dbh = shift;
 	my $sql = shift;
 	my $sth;
-	my $dsn = 'dbi:'.$dbh->{Driver}->{Name}.':'.$dbh->{Name};
+	my $dsn = $dbh->{private_dbixha_dsn} || die "panic: no private_dbixha_dsn";
 	warn "$prefix in prepare: dsn: $dsn ; sql: $sql \n" if (DBIx_HA_DEBUG > 1);
 	if (DBIx::HA::_isactivedb ($dsn)) {
-		warn "$prefix Statement handle being prepared while existing statement handle still open!\n\tdbh:\t\t$dsn\n\tprevious statement:\t".$dbh->{Statement}."\n\tcurrent statement:\t$sql\n\tACTIVE KIDS: ".$dbh->{ActiveKids}."\n" if ($dbh->{ActiveKids});
+		warn join "\n", "$prefix Statement handle being prepared while existing statement handle still open!",
+                        "\tdbh:\t\t$dsn",
+                        "\tprevious statement:\t".$dbh->{Statement},
+                        "\tcurrent statement:\t$sql",
+                        "\tACTIVE KIDS: ".$dbh->{ActiveKids}."\n"
+                    if $dbh->{ActiveKids};
 	} else {
 		my $dbname = DBIx::HA::_getdbname($dsn);
 		($dsn, $dbh) = DBIx::HA::_reconnect ($dsn, $dbh);
@@ -404,7 +433,7 @@ sub prepare {
 			return undef;
 		}
 	}
-	$sth = $dbh->SUPER::prepare($sql);
+	$sth = $dbh->SUPER::prepare($sql,@_);
 	return $sth;
 }
 	
@@ -412,6 +441,7 @@ sub prepare {
 
 {
 package DBIx::HA::st;
+use strict;
 use constant DBIx_HA_DEBUG => DBIx::HA::DBIx_HA_DEBUG;
 use Sys::SigAction qw( set_sig_handler );
 use vars qw ( @ISA $prefix );
@@ -422,11 +452,13 @@ sub execute {
 	my $sth = shift;
 	my $dbh = $sth->{Database};
 	my $sql = $dbh->{Statement};
-	my $dsn = 'dbi:'.$dbh->{Driver}->{Name}.':'.$dbh->{Name};
+	my $dsn = $dbh->{private_dbixha_dsn} || die "panic: no private_dbixha_dsn";
+	my $dbname = DBIx::HA::_getdbname($dsn);
 	my $res;
 	my $to;	# did we trip a timeout on the execution?
 	my $orig_error_code;
 	my $orig_error_string;
+	my $max_executions = $DATABASE::conf{$dbname}->{'max_retries'} * scalar(@{$DATABASE::conf{$dbname}->{'db_stack'}});
 
 	warn "=================\n" if (DBIx_HA_DEBUG > 1);
 	warn "$prefix in execute: dsn: $dsn ; sql: $sql \n" if (DBIx_HA_DEBUG > 1);
@@ -434,24 +466,13 @@ sub execute {
 		($res, $to) = &_execute_with_timeout ($dsn, $sth);
 		$orig_error_code = $DBI::err;
 		$orig_error_string = $DBI::errstr;
-		if ($to) {
-			# It was a timeout error. The database handle may be in an unstable state, we must clear it.
-			# To check if the database is still active, we try to connect to it again.
-			# If that succeeds, then the database is fine. If not, then we need to fail over.
-			my ($dummy_dsn, $username, $auth, $attrs) = @{$DATABASE::conf{&DBIx::HA::_getdbname($dsn)}->{'active_db'}};
-			my $newdbh = &DBIx::HA::_connect_with_timeout($dsn, $username, $auth, $attrs);
-			if (! $newdbh) {
-				# Ooops. Even a new database connect fails. It's time to reconnect and reexecute
-				warn "$prefix in execute: execution failed with timeout, database unavailable. Attempting reconnect (with potential failover). Statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
+		if ($to || ((! defined $res) && &{$DATABASE::conf{$dbname}->{'failtest_function'}}($orig_error_code, $orig_error_string))) {
+			# It was a timeout error or a critical network library error (connection in a bad state)
+			warn "$prefix in execute: timeout error or network lib error, reexecuting: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
+			for (my $count_execs = 0; $count_execs < $max_executions; $count_execs++) {
 				($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
-			} else {
-				# the reconnect worked, so the database is fine.
-				# get rid of the old database handle
-				warn "$prefix *** TIMEOUT! server busy: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
-				$dbh->swap_inner_handle($newdbh);
-				eval { $newdbh->disconnect; };
-				undef $newdbh;
-			}
+				last if (defined $res); # reexecution worked (or failed hard with -666)
+			} 
 		} elsif (! defined $res) {
 			# We got an error code from the server upon statement execution.
 			# We will let the client decide what to do and let it be.
@@ -459,8 +480,10 @@ sub execute {
 			warn "$prefix in execute: bad sql: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
 		}
 	} else { # current db is not active
-		eval { $sth->finish; };
-		($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
+		for (my $count_execs = 0; $count_execs < $max_executions; $count_execs++) {
+			($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
+			last if (defined $res); # reexecution worked (or failed hard with -666)
+		}
 	}
 	if (! defined $res) { # Execution failed
 		warn "$prefix in execute: result is undefined, statement execution failed! statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
@@ -468,7 +491,7 @@ sub execute {
 		return undef;
 	}
 	if ($res == -666) { # the devil killed you! We couldn't connect to the db!
-		warn "$prefix in execute: statement couldn't be executed because connect failed abnormally. statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
+		warn "$prefix in execute: statement couldn't be executed because connect failed abnormally. statement: $sql ; dsn: $dsn \n";
 		warn "+++++++++++++++++\n" if (DBIx_HA_DEBUG > 1);
 		return undef;
 	}
@@ -480,6 +503,28 @@ sub execute {
 	return $res;
 }
 
+=begin private
+
+=head2
+
+ ($execute_result,$timeout_triggered) = _execute_with_timeout($dsn,$sth);
+
+Calls "execute" on a DBI statement handle, and handles a possible timeout of the query.
+
+Args:
+ $dsn: a key in our internal lookup table of connection details
+ $sth: DBI statement handle
+
+Returns:
+  - result of execute() call 
+  - boolean, true if timeout was triggered. 
+    
+
+=end private
+
+=cut
+
+
 sub _execute_with_timeout {
 	my $dsn = shift;
 	my $sth = shift;
@@ -501,7 +546,7 @@ sub _execute_with_timeout {
 	alarm(0);
 	if ($@ or $timeout) {	# there's a problem above
 		if ($timeout) {	# it's a timeout
-			warn "$prefix EXECUTION TIMED OUT in $dsn ; SQL: $sql";
+			warn "$prefix *** EXECUTION TIMED OUT in $dsn ; SQL: $sql";
 		} else {	# problem in the execution
 			warn "$prefix Error in DBI::execute: $@\n" if $@;
 		}
@@ -515,9 +560,7 @@ sub _execute_with_timeout {
 sub _reexecute {
 	# reexecute the statement in the following way:
 	# reconnect with a new dbh
-	# redo prepare and execute
-	# only do it once, since if the connection works and the execution doesn't,
-	# it means that the statement sql is wrong and should not be retried.
+	# redo prepare and execute until it works
 	my $dsn = shift;
 	my $sql = shift;
 	my $sth = shift || undef;
@@ -532,7 +575,7 @@ sub _reexecute {
 		$dbh = $sth->{Database};
 	}
 	($dsn, $dbh) = DBIx::HA::_reconnect ($dsn, $dbh);
-	if (! defined $dbh) { return ($dsn, -666); } # we couldn't connect at all
+	if (! defined $dbh) { return ($dsn, $sth, -666); } # we couldn't connect at all
 	$newsth = $dbh->prepare($sql);
 	($res, $to) = &_execute_with_timeout ($dsn, $newsth);
 	if (! $res) {	# execute_with_timeout failed
@@ -586,6 +629,7 @@ DBIx::HA - High Availability package for DBI
     connecttimeout  => 1,
     executetimeout  => 8,
     callback_function => \&MyCallbackFunction,
+	failtest_function   => \&DBIx::HA::FTF_SybaseASE,
     };
 
  DBIx::HA->initialize();
@@ -681,15 +725,16 @@ datasource #2 if available.
 
 =item connectoninit ( DEFAULT: 0 )
 
-If set to 1, then during the I<initialize()> phase this database connections
-will be instantiated with its currently  active db_stack entry.
-This is very useful under mod_perl and replaces the C<Apache::DBI>
-I<connect_on_init()> method. 
+If set to 1 and L<Apache::DBI> has already been loaded, then during the
+I<initialize()> phase database connections will be opened with the
+currently active db_stack entry.  This is very useful under mod_perl and
+replaces the purpose of the C<Apache::DBI> I<connect_on_init()> method. 
 
 =item pingtimeout ( DEFAULT: -1 )
 
-this is only useful in conjunction with C<Apache::DBI>. The default of -1
-disables pinging the datasource. It is recommended not to modify it. See
+This configures the usage of the ping method, to validate a connection.  The
+option is only checked if L<Apache::DBI> has already been loaded. The default
+of -1 disables pinging the datasource. It is recommended not to modify it. See
 C<Apache::DBI> for more information on ping timeouts. Timeout is in seconds.
 
 =item failoverlevel ( DEFAULT: process )
@@ -707,30 +752,56 @@ could be potentially hitting a different physical database server.
 
 =item application
 
-a file-based interprocess communication is used to notify Apache/mod_perl
+A file-based interprocess communication is used to notify Apache/mod_perl
 processes of the currently active datasource. This allows all processes to fail
 over near-simultaneously. A process in the middle of an I<execute> will do it
 on the next call to I<prepare> or I<execute>. This is only available under
-mod_perl.
+mod_perl. It only has an effect if we detect that mod_perl is in effect, by
+checking that C<$Apache::VERSION> has a value.
 
 =back
 
-=item connecttimeout ( DEFAULT: 5 )
+=item connecttimeout ( DEFAULT: 30 )
 
-timeout for connecting to a datasource, in seconds.
+Timeout for connecting to a datasource, in seconds. A value of 0 disables this timeout.
 
-=item executetimeout ( DEFAULT: 10 )
+=item executetimeout ( DEFAULT: 30 )
 
-timeout for execution of a statement, in seconds. If the timeout is triggered,
+Timeout for execution of a statement, in seconds. If the timeout is triggered,
 the database handle is deleted and a new connect is tried. If the connect
 succeeds, we assume that the problem is with a runaway SQL statement or bad
-indexing. If the connect fails, then we fail over.
+indexing. If the connect fails, then we fail over. A value of 0 disables this timeout.
 
 =item callback_function ( DEFAULT: I<none> )
 
 reference to a function to call whenever the datasource is changed due to a
 failover. See the TIPS sections for a usage example.
 
+=item failtest_function ( DEFAULT: sub{0} )
+
+Reference to a function to call to test if a DBI error is a candidate for
+failover or not. This is only triggered when a call to C<execute()> returns
+an undefined value.
+
+Input is ($DBI::err, $DBI::errstr). These correspond to the native driver error
+code and string values. See the docs for your database driver and L<DBI> for
+details.
+
+Output is boolean: If true, then we'll consider the error a critical
+condition, ok to failover. If false, then DBIx::HA will not act on it
+and pass it straight through to the client.
+
+This Fail Test Function (FTF) function is extremely important for the proper
+functioning of DBIx::HA. You must be careful  in defining it
+precisely, based on the database engine that you are using. A sample
+function for Sybase is included:
+
+  failtest_function   => \&DBIx::HA::FTF_SybaseASE,
+
+To consider any error a reason to failover, you could use the following:
+
+  failtest_function   => sub {1},
+
 =back
 
 =head1 USER METHODS
@@ -839,17 +910,29 @@ bounce Apache.
 
 =head1 DEPENDENCIES
 
-This modules requires Perl >= 5.6.0, DBI >= 1.44  and Sys::SigAction.
-Apache::DBI is recommended when using mod_perl.
-If using Apache::DBI, version 0.89 or above is required.
-Always load Apache::DBI and Apache before DBIx::HA if you want DBIx::HA to know
-of them.
+This modules requires Perl >= 5.6.0, DBI >= 1.49  and Sys::SigAction.
+
+Apache::DBI is recommended when using mod_perl.  If using Apache::DBI, version 0.89 or above is required.
+Always load Apache::DBI and Apache before DBIx::HA if you want DBIx::HA to know of them.
+
+If using PostgreSQL, use DBD::Pg 2.0 or newer. Older versions of DBD::Pg contain a bug
+which make it incompatible with this module. 
 
 =head1 BUGS
 
 Currently I<%DATABASE::conf> needs to be manually and directly populated.
 A proper interface needs to be built for it.
 
+=head1 URLS
+
+The DBIx::HA project is hosted in Google Code:
+  http://code.google.com/p/perl-dbix-ha/
+
+Please submit bug reports or feature improvements requests to the site above.
+
+The svn repository is also at:
+  https://perl-dbix-ha.googlecode.com/svn/
+
 =head1 SEE ALSO
 
 C<DBD::Multiplex> for simultaneous writes to multiple data sources.
@@ -864,7 +947,7 @@ Henri Asseily <henri@asseily.com>
 
 =head1 COPYRIGHT
 
-Copyright (c) 2003-2004 Henri Asseily <henri@asseily.com>.
+Copyright (c) 2003-2006 Henri Asseily <henri@asseily.com>.
 All rights reserved. This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
 
